<?php
/**
 *
 * @package 
 * @author     Ole Jungclaussen
 * @version    0.9.0
 * @.copyright 2021, Ole Jungclaussen
 */
/**
 * A server's Info retrieved from STEAM
 * @package 
 */
class CServerBySteam {
    /**
     * 
     * @var string
     */
    public $sIp   = null;
    /**
     * 
     * @var string
     */
    public $sPort = null;
    /**
     * 
     * @param string $sIp 
     * @param string $sPort 
     * @return void 
     */
    function __construct(string $sIp, string $sPort)
    {
        $this->sIp   = $sIp;
        $this->sPort = $sPort;
    }
}
/**
 * Server List from STEAM
 * @package 
 */
class CServerlistBySteam {
// ########## PROPS PUBLIC
    /**
     * STEAM
     * @var int
     */
    const CFG_MASTERSERVER_MIN_PORT = 27011;
    /**
     * STEAM
     * @var int
     */
    const CFG_MASTERSERVER_MAX_PORT = 27012;
    /**
     * STEAM
     * @var int
     */
    const CFG_MAX_TRY_MASTER = 10;
    /**
     * STEAM
     * app id; Cliffs Of Dover Blitz
     * @var string
     */
    const CFG_SERVER_FILTER  = '\appid\754530';
    /**
     * STEAM
     * @var string
     */
    const CFG_SERVER_REGION  = "\xFF";           // region = world
    /**
     * STEAM
     * @var string
     */
    const CFG_TIMEOUT        = 1;                // seconds timeout
    /**
     * 
     * @var true
     */
    public $bPrintDiagnostics = false;
    /**
     * Servers running the game
     * 
     * @var array
     */
    public $aServers = [];
    /**
     * Number of servers running the game
     * @var int
     */
    public $nServers = 0;
// ########## PROPS PROTECTED
    /**
     * Current Time (UNIX_TIME_STAMP) of Query.
     * @var int
     */
    protected $tsNow = null;
// ########## PROPS PRIVATE
// ########## CONST/DEST
    function __construct()
    {
        
    }
    function __destruct()
    {
        
    }
// ########## METHOD PUBLIC
    /**
     * Query List of Servers.
     * 
     * @return void 
     * @throws Exception 
     */
    public function load() : void
    {
        $nTryMaster = 0;
        /// Get a list of IPv4 addresses corresponding to a given Internet host name
        $aMasterServers = gethostbynamel("hl2master.steampowered.com");

        /// Try all master servers until we find one that isn't timing out.
        do{
            $rSocket   = null;
            $bTimedOut = false;
            $nTryMaster++;
           
            foreach($aMasterServers as $ip){
                /// got a connection!
                if(false==($bTimedOut = $this->connectToMasterServer_timeout($rSocket, $ip))){
                    if($this->bPrintDiagnostics)
                    {
                        print '$bTimedOut == false'.PHP_EOL;
                    }
                    break;
                }
                /// timed out, close socket if any
                else if($rSocket){
                    fclose($rSocket);
                    $rSocket = null;
                }
            }
            
            if($bTimedOut)
            {
                throw new Exception('timeout on *all* master server['.implode(', ', $aMasterServers).']');
            }

            /// Read list with server addresses (IP:port).
            $this->nServers = 0;
            $PrevByte1 = 0;
            $PrevByte2 = 0;
            $PrevByte3 = 0;
            $PrevByte4 = 0;
            $PrevInt5  = 0;

            do {
                stream_set_timeout($rSocket, self::CFG_TIMEOUT);

                /// IP address: "$a1.$a2.$a3.$a4"
                /// Port:       "$a5"
                $Byte1 = ord(fread($rSocket,1));
                $info = stream_get_meta_data($rSocket);
    
                if ($info['timed_out']){
                    $seed = "$PrevByte1.$PrevByte2.$PrevByte3.$PrevByte4:$PrevInt5";
                    
                    while($this->query_timeout($rSocket, $seed));
                    
                    stream_set_timeout($rSocket, self::CFG_TIMEOUT);
                    $Byte1 = ord(fread($rSocket,1));
                    $info = stream_get_meta_data($rSocket);
                    if ($info['timed_out']){
                        break;
                    }
                    $check_for_duplicate = 1;
                }
                else $check_for_duplicate = 0;
    
                /// Always read the rest of the address (even if it starts with 0) in
                /// order to empty the master server's write buffer.
                $Byte2 = ord(fread($rSocket, 1));
                $Byte3 = ord(fread($rSocket, 1));
                $Byte4 = ord(fread($rSocket, 1));
                $Int5 = ord(fread($rSocket, 1))*256 + ord(fread($rSocket, 1));
    
                if($Byte1 != 0){
                    if(
                        ($check_for_duplicate==0)
                        || ($Byte1!=$PrevByte1)
                        || ($Byte2!=$PrevByte2)
                        || ($Byte3!=$PrevByte3)
                        || ($Byte4!=$PrevByte4)
                        || ($Int5!=$PrevInt5)
                    ){
                        try {
                            $this->aServers[] = new CServerBySteam("$Byte1.$Byte2.$Byte3.$Byte4", "$Int5");
                            $this->nServers++;
                        }
                        catch(Exception $e){
                            /// We choose to ignore errors, here
                        }
                    }
                    $PrevByte1 = $Byte1; $PrevByte2 = $Byte2; $PrevByte3 = $Byte3; $PrevByte4 = $Byte4; $PrevInt5 = $Int5;
                }
            } while($Byte1 != 0);
            
            fclose($rSocket);

        } while(0==$this->nServers && $nTryMaster <= self::CFG_MAX_TRY_MASTER);
    }
    /**
     * 
     * @return int 
     */
    public function countServers() : int
    {
        return($this->nServers);
    }
    /**
     * 
     * @return int
     */
    public function countPlayers() : int
    {
        $nCount = 0;
        for($i=0; $i < $this->nServers; $i++)
        {
            $nCount += $this->aServers[$i]->countPlayers();
        }
        return($nCount);
    }
    /**
     * 
     * @return void 
     */
    public function destroy() : void
    {
    }
// ########## METHOD PROTECTED
    /**
     * Connect to master server.
     * The socket is passed as reference and thus returned as well.
     * 
     * @param resource $rSocket 
     * @param string $ip 
     * @return bool true on timeout
     */
    protected function connectToMasterServer_timeout(&$rSocket, string $ip) : bool
    {
        $iPort    = self::CFG_MASTERSERVER_MIN_PORT;
        $bTimeout = false;
        do {
            if(true==$this->bPrintDiagnostics)
            {
                print "connectToMasterServer_timeout: ".$ip."<br>";
            }
            $rSocket = fsockopen("udp://$ip", $iPort, $errno, $errstr, self::CFG_TIMEOUT);
            
            if(false===$rSocket)
            {
                if(true==$this->bPrintDiagnostics)
                {
                    print "connectToMasterServer_timeout: No socket<br>";
                }
                return true;
            }
            $bTimeout = $this->query_timeout($rSocket, "0.0.0.0:0");
            $iPort = $iPort + 1;
            
        } while($bTimeout && ($iPort <= self::CFG_MASTERSERVER_MAX_PORT));
        
        return $bTimeout;
    }
    /**
     * Query socket.
     * 
     * @param resource $socket 
     * @param string $seed 
     * @return bool true on timeout
     */
    protected function query_timeout(&$socket, string $seed) : bool
    {
        stream_set_timeout($socket, self::CFG_TIMEOUT);

        if(!fwrite($socket, "1".self::CFG_SERVER_REGION."$seed\0".self::CFG_SERVER_FILTER."\0")){
            fclose($socket);
            return true;
        }
    
        stream_set_timeout($socket, self::CFG_TIMEOUT);
        $s = bin2hex(fread($socket, 6));
        $info = stream_get_meta_data($socket);
        if(true==$this->bPrintDiagnostics)
        {
            print "query_timeout Read[".$s."]";
            foreach($info as $k => $v) print $k."[".$v."]";
            print "<br>";
        }

        if($info['timed_out']);
        else if($s !== "ffffffff660a"){
            fclose($socket);
            return true;
        }
        
        return $info['timed_out'];
    }
// ########## METHOD PRIVATE
};
